-- Widget pooling and stuff

local METHODS = {};
local SET_SCRIPT_REPLACEMENTS = {};

local function GetReplacementSetScript(oldFunction)
   local newFunc = SET_SCRIPT_REPLACEMENTS[oldFunction];
   if (newFunc) then
      return newFunc;
   end
   
   newFunc = function(self, handler, func)
		local oldSet = self.oldHandlers;
		if (oldSet == nil) then
		   oldSet = {};
		   self.oldHandlers = oldSet;
		end
		local oldH = oldSet[handler];
		if (oldH == false) then
		   oldH = nil;
		elseif (oldH == nil) then
		   oldH = false;
		end
		if (func == oldH) then
		   -- Restoring old handler
		   oldSet[handler] = nil;
		   return oldFunction(self, handler, func);
		end
		if (oldH == false) then
		   local prevH = self:GetScript(handler);
		   if (prevH == func) then
		      return;
		   end
		   if (oldH == false) then
		      oldSet[handler] = prevH or false;
		   end
		end
		return oldFunction(self, handler, func);
	     end
   SET_SCRIPT_REPLACEMENTS[oldFunction] = newFunc;
   -- Paranoia
   SET_SCRIPT_REPLACEMENTS[newFunc] = newFunc;
   return newFunc;
end

function StatRingsOptions_RegisterWidget(widget, type, owner)
   if (not owner) then
      owner = widget:GetParent();
   end
   local widgets = owner.widgets;
   if (not widgets) then
      widgets = {};
      owner.widgets = widgets;
   end
   local wtype = widgets[type];
   if (not wtype) then
      wtype = {};
      widgets[type] = wtype;
   end
   table.insert(wtype, widget);
   
   widget.SetScript = GetReplacementSetScript(widget.SetScript);
end

function StatRingsOptions_WidgetManager(frame)
   local widgets = frame.widgets;
   if (not widgets) then
      widgets = {};
      frame.widgets = widgets;
   end

   local active = frame.activeWidgets;
   if (not active) then
      active = {};
      frame.activeWidgets = active;
   end

   -- Apply methods
   for k,v in pairs(METHODS) do
      frame[k] = v;
   end
end

function METHODS:ReleaseAllWidgets()
   local active = self.activeWidgets;
   for k,v in pairs(active) do
      k:Hide();
      if (v.Restore) then
	 v.Restore(v,k);
      end
      k.widgetInfo = nil;
      active[k] = nil;
      local oldHandlers = k.oldHandlers;
      if (oldHandlers) then
	 DEFAULT_CHAT_FRAME:AddMessage("Restoring handlers for " 
				       .. k:GetName());
	 for h,v in pairs(oldHandlers) do
	    k:SetScript(h, v or nil);
	 end
	 k.oldHandlers = nil;
      end
   end
end

function METHODS:AllocateWidget(widgetInfo, defaultSetup)
   local widgets = self.widgets;
   local active = self.activeWidgets;
   local wtypes = widgets[widgetInfo.type];
   if (not wtypes) then
      return nil;
   end
   for _,v in ipairs(wtypes) do
      if (not active[v]) then
	 v.widgetInfo = widgetInfo;
	 active[v] = widgetInfo;
	 local setup = widgetInfo.Setup or defaultSetup;
	 if (setup) then
	    setup(widgetInfo, v);
	 end
	 return v;
      end
   end
   for k,v in active do
      k:Hide();
      active[k] = nil;
   end
end

function METHODS:ActiveWidgetIterator()
   return pairs(self.activeWidgets);
end